package ru.dbotthepony.kommons.random

import java.util.random.RandomGenerator

/**
 * Excellent PRNG with expected period of 2^255 - 1, a slightly faster version than [Xoshiro256StarStarRandom].
 *
 * Where [Xoshiro256PlusPlusRandom] is used consider using [Xoshiro256StarStarRandom] instead.
 */
open class Xoshiro256PlusPlusRandom protected constructor(
	@JvmField
	protected var s0: Long,
	@JvmField
	protected var s1: Long,
	@JvmField
	protected var s2: Long,
	@JvmField
	protected var s3: Long,
	mark: Nothing?
) : RandomGenerator {
	init {
		if (s0 or s1 or s2 or s3 == 0L) {
			s0 = 0x73CF3D83FFF44FF3L
			s1 = 0x6412312B70F3CD37L
			s2 = -0X6BB4C4E1327BFDCFL
			s3 = -0X4BE0F5BB5F3F5240L
		}
	}

	constructor(s0: Long, s1: Long, s2: Long, s3: Long) : this(s0, s1, s2, s3, null) {
		// discard some values so generator can get going if provided seed was "weak"
		for (i in 0 until 32)
			round()
	}

	constructor(seed: Long) : this(1L, 0L, 0L, 0L, null) {
		val rng = PCG32Random(seed)
		s0 = rng.nextLong()
		s1 = rng.nextLong()
		s2 = rng.nextLong()
		s3 = rng.nextLong()
	}

	protected fun round(): Long {
		val result = (s0 + s3).rotateLeft(23) + s1
		val t = s1.shl(17)
		s2 = s2.xor(s0)
		s3 = s3.xor(s1)
		s1 = s1.xor(s2)
		s0 = s0.xor(s3)
		s2 = s2.xor(t)
		s3 = s3.rotateLeft(45)
		return result
	}

	override fun nextLong(): Long {
		return round()
	}

	companion object {
		@JvmStatic
		fun raw(
			s0: Long,
			s1: Long,
			s2: Long,
			s3: Long,
		): Xoshiro256PlusPlusRandom {
			return Xoshiro256PlusPlusRandom(s0, s1, s2, s3, null)
		}
	}
}
